/* -*- Mode: Text; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Chatzilla.
 *
 * The Initial Developer of the Original Code is
 * New Dimensions Consulting, Inc.
 * Portions created by the Initial Developer are Copyright (C) 1999
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Robert Ginda, rginda@ndcico.com, original author
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/* 
 *  a version of the output window that uses only your default foreground and
 *  backgrond colors.  Message types are distinguished by font changes.
 *  see output-base.css for details. 
 */

@import url(chrome://chatzilla/content/output-base.css);

.header-outer {
    background-color: white;
}

.progress-fg {
    background-color: grey;
}

.msg[msg-type="ACTION"] .msg-data {
    font-style: italic;
}

.msg[msg-type="JOIN"] .msg-type,
.msg[msg-type="PART"] .msg-type,
.msg[msg-type="QUIT"] .msg-type {
    font-weight: bold;
}

.msg[msg-type="QUIT"] .msg-data {
    font-variant: small-caps;
    font-weight: bold;
}

.msg[msg-type="JOIN"] .msg-data,
.msg[msg-type="PART"] .msg-data {
    font-variant: small-caps;
}

.msg[msg-type="HELLO"] .msg-data,
.msg[msg-type="NICK"] .msg-type,
.msg[msg-type="NOTICE"] .msg-data {
    font-weight: bold;
}

.msg[msg-type="NICK"] .msg-data {
    font-family: monospace;
}

/* :before and :after pseudoclasses form the decorations around nicknames.     */
.msg-user:before,
.msg-user:after {
    font-size: 100%;
    font-family: monospace;
    font-weight: bolder;
}

.msg[dest-type="IRCUser"] .msg-user,
.msg[dest-type="IRCUser"][msg-dest$="ME!"] .msg-user {
    font-style: italic;
}

.msg[msg-user$="ME!"] .msg-user {
    font-weight: bold;
}
