/*
 *  This defines the look-and-feel styling of the error pages.
 *  (see: netError.xhtml)
 *
 *  Original styling by William Price <bugzilla@mob.rice.edu>
 *  Updated by: Steven Garrity <steven@silverorange.com>
 *              Henrik Skupin  <mozilla@hskupin.info>
 *              Jon Hicks <jon@hicksdesign.co.uk>
 *              Philippe Wittenbergh <phiw@l-c-n.com>
 *              Sean Murphy <murph@seanmurph.com> (Blacklist styling)
 */

html {
  background: #F2F2F2;
}

body {
  margin: 0;
  padding: 0 1em;
  color: -moz-FieldText;
  font: message-box;
}

h1 {
  margin: 0 0 .3em 0;
  font-size: 240%;
  color: #000;
  font-weight: normal;
}

ul, ol {
  margin: 0 0 1em 0;
  -moz-margin-start: 1.5em;
  padding: 0;
}

*|*::-moz-list-bullet {
  color: #aaa;
  font-size: 1.2em
}

ul > li, ol > li {
  margin-bottom: 0.7em;
}

ul {
  list-style: disc;
}

#errorPageContainer {
  position: relative;
  min-width: 13em;
  max-width: 52em;
  margin: 4em auto;
  border: 2px solid #dadada;
  -moz-border-radius: 15px;
  padding: 3em;
  -moz-padding-start: 60px;
  background: url("chrome://global/skin/icons/warning.png") 1.1em 1.6em no-repeat -moz-Field;
}

/* #errorPageContainer.certerror {
  background-image: url("chrome://global/skin/icons/sslWarning.png");
}
*/

body[dir="rtl"] #errorPageContainer {
  background-position: right 0;
}

#errorTitle {
  -moz-margin-start: 64px;
}

#errorLongContent {
  -moz-margin-start: 64px;
}

#errorShortDesc > p {
  overflow: auto;
  border-bottom: 1px solid ThreeDLightShadow;
  padding-bottom: 1em;
  font-size: 130%;
  color: #0141b7;
  white-space: pre-wrap;
}

#errorLongDesc {
  -moz-padding-end: 3em;
  font-size: 110%;
}

#errorLongDesc > p {
}

#errorTryAgain {
  margin-top: 2em;
  -moz-margin-start: 64px;
}

#brand {
  position: absolute;
  right: 0;
  bottom: -1.5em;
  -moz-margin-end: 10px;
  opacity: .4;
}

body[dir="rtl"] #brand {
  right: auto;
  left: 0;
}

#brand > p {
  margin: 0;
}

#errorContainer {
  display: none;
}

#securityOverrideDiv {
  padding-top: 10px;
}

#securityOverrideContent {
  background-color: #ffdd7f; /* pale orange */
  padding: 1.3em 1.5em;
  border: 1px solid transparent; /* bug 505946 */
  -moz-border-radius: 15px;
  -moz-margin-end: 3em;
}

/* Ensure all text is left-aligned to the same position - bug 451131 comment 9 */
/* the 'link' */
#securityOverrideLink {
  -moz-margin-start: 1.5em;
}

#securityOverrideContent > p {
  margin: 0 0 1em;
}

#securityOverrideContent > button {
  margin: 0;
}

/* Custom styling for 'blacklist' error class - bug 380932  */
:root.blacklist #errorTitle, :root.blacklist #errorLongContent,
:root.blacklist #errorShortDesc, :root.blacklist #errorLongDesc {
  background-color: #e8e8e8;
  color: black;
}

:root.blacklist #errorPageContainer {
  background-image: url("chrome://global/skin/icons/blacklist.png");
  background-color: #e8e8e8;
  border: 2px solid black;
}

:root.blacklist {
  background: #722;
}

:root.blacklist #errorTryAgain {
  display: none;
}

:root.blacklist #errorShortDesc > p {
  color: black;
  font-weight: bold;
  border-bottom: 1px solid black;
}

:root.blacklist #errorTitle h1 {
  color: #722;
}

:root.blacklist #buttons {
  padding-top: 10px;
}

ul strong {
	color: #000;
}
