#define COMMENT_1 DO NOT EDIT THE VERSION IN THIS FILE! Change the version in Release/Makefile instead, then run make version while in the Release directory.
#define COMMENT_2 In case you're wondering why we're using #define for comments: This header is #imported into both Objective-C code and XML Info.plist files.

#define GROWL_VERSION 1.2
#define GROWL_VERSION_STRING "1.2"

#undef COMMENT_1
#undef COMMENT_2
