/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TransforMiiX XSLT processor.
 *
 * The Initial Developer of the Original Code is
 * IBM Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2002
 * IBM Corporation. All Rights Reserved.
 *
 * Contributor(s):
 *   IBM Corporation
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "txExpr.h"
#include "txExprResult.h"
#include "txSingleNodeContext.h"

txPredicatedNodeTest::txPredicatedNodeTest(txNodeTest* aNodeTest,
                                           Expr* aPredicate)
    : mNodeTest(aNodeTest),
      mPredicate(aPredicate)
{
    NS_ASSERTION(!mPredicate->isSensitiveTo(Expr::NODESET_CONTEXT),
                 "predicate must not be context-nodeset-sensitive");
}

PRBool
txPredicatedNodeTest::matches(const txXPathNode& aNode,
                              txIMatchContext* aContext)
{
    if (!mNodeTest->matches(aNode, aContext)) {
        return PR_FALSE;
    }

    txSingleNodeContext context(aNode, aContext);
    nsRefPtr<txAExprResult> res;
    nsresult rv = mPredicate->evaluate(&context, getter_AddRefs(res));
    NS_ENSURE_SUCCESS(rv, PR_FALSE);

    return res->booleanValue();
}

double
txPredicatedNodeTest::getDefaultPriority()
{
    return 0.5;
}

PRBool
txPredicatedNodeTest::isSensitiveTo(Expr::ContextSensitivity aContext)
{
    return mNodeTest->isSensitiveTo(aContext) ||
           mPredicate->isSensitiveTo(aContext);
}

#ifdef TX_TO_STRING
void
txPredicatedNodeTest::toString(nsAString& aDest)
{
    mNodeTest->toString(aDest);
    aDest.Append(PRUnichar('['));
    mPredicate->toString(aDest);
    aDest.Append(PRUnichar(']'));
}
#endif
