/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Simon Fraser <smfr@smfr.org>.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Simon Fraser <smfr@smfr.org> (Original Author)
 *   Smokey Ardisson <alqahira@ardisson.org>
 *   Philippe Wittenbergh <phiw@l-c-n.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
 * CSS rules to hide various ad-related images and Flash files,
 * while avoiding too many false hits.
 * 
 * Based on CSS by Joe Francis at http://www.floppymoose.com.
 * 
 */

a:link[href*="/ad."] img,
a:link[href*="/ads."] img,
a:link[href*="/adclick/"] img,
a:link[href*="/phpads/"] img, 
a:link[href*="?banner"] img,
a:link[href*="clickurl="] img,
a:link[href*=".atwola."] img,
a:link[href*="spinbox."] img,
a:link[href*="transfer.go"] img,
a:link[href*="adfarm"] img,
a:link[href*="adserver"] img,
a:link[href*=".falkag."] img,
a:link[href*="rightmedia.net"] img,
a:link[href*="advertising.com"] img,
a:link[href*="mondomedia.com"] img,
a:link[href*="mediaplex.com"] img,
a:link[href*="bluestreak"] img,
a:link[href*="doubleclick"] img,
a:link[href*="2mdn.net"] img,
a:link[href*="atdmt.com"] img,
a:link[href*="serving-sys.com"] img,
a:link[href*="linksynergy.com"] img,
a:link[href*="qksrv.net"] img,
a:link[href*="tkqlhce.com"] img,
a:link[href*="tribalfusion"] img,
a:link[href*="valueclick"] img,
a:link[href*=".fastclick.net"] img,
a:link[href*="/0AD"] img,
a:link[href*="trackoffer."] img,
a:link[href*="tracksponsor."] img,
a:link[href*="print2webcorp.com"] img,
a:link[href*="googlesyndication.com"] img,
a:link[href*="adquest3d.com"] img,
a:link[href*="/RealMedia/ads/"] img,
a:link[href*=".jdoqocy.com"] img,
a:link[href*=".zedo.com"] img,
a:link[href*="/banman.asp?"] img,
a:link[href*="adsys.townnews.com"] img,
a:link[href*="ak.m3.net"] img,
a:link[href*=".pointroll.com"] img,
a:link[href*=".mspaceads.com"] img,
a:link[href*="247realmedia.com"] img,
a:link[href*="adserv.quality-channel.de"] img,
a:link[href*="blogads.com"] img, 
a:link[href*="interclick.com"] img, 
a:link[href*="anad.tacoda.net/ads"] img,
a:link[href*="sw.interpolls.com"] img,
a:link[href*="kliptracker.com"] img,
a:link[href*="klipmart.com"] img,
a:link[href*="gfx.dvlabs.com"] img,
a:link[href*="eyereturn.com"] img,
a:link[href*="save.TV"] img,
a:link[href*="exclick."] img,
a:link[href*="adclick."] img,
a:link[href*="clickbank.net"] img,
a:link[href*=".ru4.com"] img,
a:link[href*=".burstnet.com"] img,
a:link[href*="adpeeps"] img,
a:link[href*="adbrite"] img,
a:link[class*="adText"], 
a:link[class*="adHeadline"],
a:link[href*="hrtrak.net"] img,
a:link[href*="specificclick.net"] img,
a:link[href*="adremote.timeinc.net"] img,
a:link[href*=".inetinteractive.com/"] img,
a:link[href*="adgardener.com"] img,
a:link[href*="ad.adtoma.com"] img,
a:link[href*="adimp.excite.co.jp"] img,
a:link[href*="kestrel.ospreymedialp.com"] img,
a:link[href*="netavenir."] img,
a:link[href*="euroclick.com"] img,
a:link[href^="http://advert."] img,
a:link[href*="adv_"] img,
a:link[href*=".adshuffle.com"] img,
a:link[href*=".linkstorm.net"] img,
a:link[href*=".mediaforceads.com"] img,

embed[src*="/us.yimg.com/a/"],
embed[src*="/ai.yimg.jp/"],
embed[src*="/ah.yimg.jp/bdv"],
embed[type="application/x-shockwave-flash"][src*="/ad."],
embed[type="application/x-shockwave-flash"][src*="/ads."],
embed[type="application/x-shockwave-flash"][src*="/ad/"],
embed[type="application/x-shockwave-flash"][src*="/ads/"],
embed[type="application/x-shockwave-flash"][src*="/Ads/"],
embed[type="application/x-shockwave-flash"][src*="/advert/"],
embed[type="application/x-shockwave-flash"][src*=".atwola.com"],
embed[type="application/x-shockwave-flash"][src*="mediaplex.com"],
embed[type="application/x-shockwave-flash"][src*="media.adlegend.com"],
embed[type="application/x-shockwave-flash"][src*="bluestreak"],
embed[type="application/x-shockwave-flash"][src*="adtech.de"],
embed[type="application/x-shockwave-flash"][src*=".falkag."],
embed[type="application/x-shockwave-flash"][src*="adt.m7z.net"],
embed[type="application/x-shockwave-flash"][src^="http://ad."][src*="doubleclick.net"],
embed[type="application/x-shockwave-flash"][src*="2mdn.net"],
embed[type="application/x-shockwave-flash"][src*=".fastclick.net"],
embed[type="application/x-shockwave-flash"][src*="serving-sys.com"],
embed[type="application/x-shockwave-flash"][src*="atdmt.com"],
embed[type="application/x-shockwave-flash"][src*="linksynergy.com"],
embed[type="application/x-shockwave-flash"][src*="adsrv"],
embed[type="application/x-shockwave-flash"][src*="adserver"],
embed[type="application/x-shockwave-flash"][src*="adclick"],
embed[type="application/x-shockwave-flash"][src*="adsfile."],
embed[type="application/x-shockwave-flash"][src*="adfshow?"],
embed[type="application/x-shockwave-flash"][src*="adimg"],
embed[type="application/x-shockwave-flash"][src*="/webad/"],
embed[type="application/x-shockwave-flash"][src*="?clickthru"],
embed[type="application/x-shockwave-flash"][src*="/RealMedia/ads/"],
embed[type="application/x-shockwave-flash"][src*=".advertising.com"],
embed[type="application/x-shockwave-flash"][src*=".adbutler.de"],
embed[type="application/x-shockwave-flash"][src*="ad.tomshardware.com"],
embed[type="application/x-shockwave-flash"][src*="ad4.sina.com.cn"],
embed[type="application/x-shockwave-flash"][src*=".specificmedia.com"],
embed[type="application/x-shockwave-flash"][src^="http://cdn"][src*=".eyewonder.com"],
embed[type="application/x-shockwave-flash"][src*=".tribalfusion.com/media/"],
embed[type="application/x-shockwave-flash"][src*="/flash/promotions/"],
embed[type="application/x-shockwave-flash"][src*=".adtrix.com"],
embed[type="application/x-shockwave-flash"][src*=".netshelter.net"],
embed[type="application/x-shockwave-flash"][src*=".gms1.net"],
embed[type="application/x-shockwave-flash"][src*="ak.m3.net"],
embed[type="application/x-shockwave-flash"][src*="ads.t-online.de"],
embed[type="application/x-shockwave-flash"][src*=".pointroll.com"],
embed[type="application/x-shockwave-flash"][src*="gfx.dvlabs.com"],
embed[type="application/x-shockwave-flash"][src*="gfx.klipmart.com"],
embed[type="application/x-shockwave-flash"][src*="adserv.quality-channel.de"],
embed[type="application/x-shockwave-flash"][src*=".yieldmanager.com"],
object[type="application/x-shockwave-flash"][data*=".yieldmanager.com"],
object[type="application/x-shockwave-flash"][data*="content.valencemedia.com/"],
object[type="application/x-shockwave-flash"][id*="adTower"],
object[type="application/x-shockwave-flash"][id*="728x90"],
object[type="application/x-shockwave-flash"][id*="171x600"],
object[type="application/x-shockwave-flash"][id*="400x250"],
object[type="application/x-shockwave-flash"][id*="160x600"],
object[type="application/x-shockwave-flash"][id*="300x250"],
object[type="application/x-shockwave-flash"][id*="adbox"],
object[type="application/x-shockwave-flash"][id*="perma_flash"],
object[type="application/x-shockwave-flash"][id*="skyscraper"],
embed[type="application/x-shockwave-flash"][id="Advertisement"],
embed[type="application/x-shockwave-flash"][src*="eyereturn.com"],
embed[type="application/x-shockwave-flash"][src*="media.nyadmcncserve"],
embed[type="application/x-shockwave-flash"][src*="adinterax.com"],
embed[type="application/x-shockwave-flash"][src*="adbureau.net"],
embed[type="application/x-shockwave-flash"][src*="yceml.net"],
embed[type="application/x-shockwave-flash"][src*="adsys.townnews.com"],
embed[type="application/x-shockwave-flash"][src*="images.anandtech.com/banners"],
embed[type="application/x-shockwave-flash"][src*="precisionclick.com"],
embed[type="application/x-shockwave-flash"][src*="adsatt.espn.go.com"],
embed[type="application/x-shockwave-flash"][src*="blogsmithmedia.com"],
embed[type="application/x-shockwave-flash"][src*=".ru4.com"],
embed[type="application/x-shockwave-flash"][src*=".adjuggler.com"],
embed[type="application/x-shockwave-flash"][src*="lygo.com/ly/video/ad/"],
embed[type="application/x-shockwave-flash"][src*="hrtrak.net"],
embed[type="application/x-shockwave-flash"][src*="adbrite"],
embed[type="application/x-shockwave-flash"][src*=".cachefly.net"],
embed[type="application/x-shockwave-flash"][src*=".unicast.com/"],
embed[type="application/x-shockwave-flash"][src*="content.yieldmanager.edgesuite.net"],
embed[type="application/x-shockwave-flash"][src*=".yieldmanager.net"],
embed[type="application/x-shockwave-flash"][src^="http://adinterax."],
embed[type="application/x-shockwave-flash"][src*="pubs.lemonde.fr"],
embed[type="application/x-shockwave-flash"][src*="timeinc.net/sponsors"],
embed[type="application/x-shockwave-flash"][src*=".admission.net"],
embed[type="application/x-shockwave-flash"][src*="impact-ad.jp"],
embed[type="application/x-shockwave-flash"][src*=".zedo.com"],
embed[type="application/x-shockwave-flash"][src*="harvest.adgardener.com"],
embed[type="application/x-shockwave-flash"][src="gcirm.dmp.gcion.com"],
embed[type="application/x-shockwave-flash"][src^="http://richmedia"][src*=".yimg.com"],
embed[type="application/x-shockwave-flash"][src*="cstatic.weborama.fr/weborama/adperf/"],
embed[type="application/x-shockwave-flash"][src*="l3static.weborama.fr/adperf"],
embed[type="application/x-shockwave-flash"][src*="adtech.panthercustomer.com"],
embed[type="application/x-shockwave-flash"][src*=".trafficmp.com"],
object[type="application/x-shockwave-flash"][src*="ad.advertstream.com"],
embed[type="application/x-shockwave-flash"][src*="2mdn.aolcdn.com"],
embed[type="application/x-shockwave-flash"][src*="connextra.com"],
embed[type="application/x-shockwave-flash"][src*="ad.adtoma.com"],
embed[type="application/x-shockwave-flash"][src*="e-planning.net"],
embed[type="application/x-shockwave-flash"][src*=".tradedoubler.com"],
embed[type="application/x-shockwave-flash"][src*="netavenir."],
embed[type="application/x-shockwave-flash"][src*="blogbang.com"],
embed[type="application/x-shockwave-flash"][src*="653.studyrama.com"],
embed[type="application/x-shockwave-flash"][src*="hs.interpolls.com"][src*="ipoll_ad.swf"],
embed[type="application/x-shockwave-flash"][src*="adtechus.com"],
embed[type="application/x-shockwave-flash"][src*="fmpub.net"],
embed[type="application/x-shockwave-flash"][src*="splastk.com"],
embed[type="application/x-shockwave-flash"][src*="ads.prisa.com"],
embed[type="application/x-shockwave-flash"][src*="media.contextweb.com"],
embed[type="application/x-shockwave-flash"][src*="adform.net"],
embed[type="application/x-shockwave-flash"][src*="ads.mixi.jp"],
embed[type="application/x-shockwave-flash"][src*=".optmd.com"],
embed[type="application/x-shockwave-flash"][src*=".afy11.net"],
embed[type="application/x-shockwave-flash"][src*="xx.openx.com"],
embed[type="application/x-shockwave-flash"][src*="adbureau.net"],
embed[type="application/x-shockwave-flash"][src*="adreactor.com/services/advertisements/"],
object[type="application/x-shockwave-flash"][data*="server.cpmstar.com"],

/* Flash trackers */
object[data*=".clearspring.com"],
embed[src*="secureinclude.ebaystatic.com"][src$="krb.swf"],
object[data*="flookies07.swf"],

img[src*="googlesyndication.com"],
img[src*=".doubleclick.net"],
img[src*="mediaplex.com"],
img[src*="serving-sys.com"],
img[src*="us.yimg.com/a/"],
img[src*="/ai.yimg.jp/"],
img[src*="/ah.yimg.jp/bdv"],
img[src*="/side_ads/"],
img[src*="/ads/"],
img[src*="/Ads/"],
img[src*="/ad/"],
img[src*="/adx/"],
img[src*="/adpics/"],
img[src*="/ads."],
img[src*="/adimages/"],
img[src*="adsfile."],
img[src*="/adz/"],
img[src*="Editorial_Promos"],
img[src*="/toms_skyscraper_"],
img[src*="bannerad_"],
img[src*="ad4.sina.com.cn"],
img[src*=".ultramercial.com"],
img[src*=".casalemedia.com"],
img[src*="2mdn.net"],
img[src*=".zedo.com"],
img[src*=".mspaceads.com"],
img[src*=".adbrite.com"],
img[src*="ximages.offeroptimizer.com"],
img[src*="kermit.macnn.com"],
img[src*="media.adlegend.com"],
img[src*=".adtrix.com"],
img[src*=".netshelter.net"],
img[src*=".fastclick.net"],
img[src*=".gms1.net"],
img[src*="ak.m3.net"],
img[src*="atdmt.com"],
img[src*=".falkag."],
img[src*=".pointroll.com"],
img[src*="gfx.dvlabs.com"],
img[id*="km_placer"],
img[src*="247realmedia.com"],
img[src*="specificclick.net"],
img[src*="admarketplace.net"],
img[src*="kliptracker.com"],
img[src*="klipmart.com"],
img[src*="gfx.dvlabs.com"],
img[src*="eyereturn.com"],
img[src*="alphaload.de/img/banner"],
img[src*="www.save.tv/STVAF"],
img[src*="tradedoubler.com"],
img[src*="yceml.net"],
img[src*="awltovhc.com"],
img[src*="adjuggler.com"],
img[src*="sharesale.com"],
img[src*="content.ipro.com/images"],
img[src*="static.fmpub.net/banners"],
img[src*="content.cpxinteractive.com"],
img[src*="righthandads"],
img[src*=".ru4.com"],
img[src*=".burstnet.com"],
img[src*="/images/advertising/"],
img[src*="hrtrak.net"],
img[src*=".tribalfusion.com"],
img[src*="cdn.specificmedia.com"],
img[src*="content.yieldmanager.edgesuite.net"],
img[src*=".yieldmanager.com"],
img[src*=".yieldmanager.net"],
img[src*="phpAds"],
img[src*=".advertserve.com/advertpro"],
img[src*="juggler.inetinteractive.com/banners/"],
img[src*="euroclick.com"],
img[src*="show.onenetworkdirect.net"],
img[src*="impact-ad.jp"],
img[src*="/RealMedia/ads/"],
img[src*="mercuras.com"],
img[src*="images.industrybrains.com/banner"],
img[src*="home.edt02.net/emc/banner/"],
img[src*="connextra.com"],
img[src*="ad.adtoma.com"],
img[src*="openx.foxnews.com"],
img[src*="netavenir."],
img[src*="blogbang.com"],
img[src*="adtechus.com"],
img[src*="ftjcfx.com"],
img[src*="lightningcast.com"],
img[src*="trafficmp.com"],
img[src*=".advertising.com"],
img[src*="/ads2.msn.com"],
img[src*=".adsdk.com"],
img[src*=".adbureau.net"],
img[src*=".adshuffle.com"],
img[src*="media.adfrontiers.com"],
img[src*="ads.mixi.jp"],
img[src*="/openads/"],
img[src*=".linkstorm.net"],
img[src*=".optmd.com"],
img[src*="xx.openx.com"],
img[src*="buysellads.com"],

iframe[src*="/ad."],
iframe[src*="/ads."],
iframe[src*="/ad/"],
iframe[src*="/ads/"],
iframe[src*="/ads2/"],
iframe[src*="/AdServer/"],
iframe[src*="/jsads/"],
iframe[src*="/ad_data/"],
iframe[src*="?banner"],
iframe[src*="_Banner"],
iframe[src*="BET365"],
iframe[src*="=click"],
iframe[src*="clickurl="],
iframe[src*="?adpartner"],
iframe[src*="googlesyndication.com"],
iframe[src*="google/box"],
iframe[src*=".atwola."],
iframe[src*="spinbox."],
iframe[src*="transfer.go"],
iframe[src*="adfarm"],
iframe[src*="adserv"],
iframe[src*=".falkag."],
iframe[src*="adt.m7z.net"],
iframe[src*="rightmedia.net"],
iframe[src*="advertising.com"],
iframe[src*="atdmt.com"],
iframe[src*="linksynergy.com"],
iframe[src*="mondomedia.com"],
iframe[src*="mediaplex.com"],
iframe[src*="media.adlegend.com"],
iframe[src*="bluestreak"],
iframe[src*="doubleclick"],
iframe[src*="casalemedia.com"],
iframe[src*="2mdn.net"],
iframe[src*="serving-sys.com"],
iframe[src*="qksrv.net"],
iframe[src*="tribalfusion"],
iframe[src*="valueclick"],
iframe[src*=".fastclick.net"],
iframe[src*="/rd."],
iframe[src*="/0ad"],
iframe[src*="trackoffer."],
iframe[src*="tracksponsor."],
iframe[src*="bridgetrack.com"],
iframe[src*=".trafficmp.com"],
iframe[src*=".yieldmanager.com"],
iframe[src*=".yieldmanager.net"],
iframe[src*=".mspaceads.com"],
iframe[src*=".adtrix.com"],
iframe[src*=".netshelter.net"],
iframe[src*=".gms1.net"],
iframe[src*="ak.m3.net"],
iframe[src*="adsvr.adknowledge.com"],
iframe[src*=".pointroll.com"],
iframe[id^="google_ads_frame"],
iframe[src*="lycos.spiegel.de"],
iframe[src*="matrix.mediavantage.de"],
iframe[src^="http://de"][src*=".myspace.com"],
iframe[src*="adjuggler.com"],
iframe[src*="clickhype.com"],
iframe[src*="euroclick.com"],
iframe[src*="network.realmedia.com"],
iframe[src*="adcom"],
iframe[src*="cpvtext.cpvfeed.com"],
iframe[src*="ypn-js.overture.com"],
iframe[src*="yceml.net"],
iframe[src*="adsys.townnews.com"],
iframe[src*="media.adrevolver.com"],
iframe[src*="kanoodle"],
iframe[src*="nextag.com"],
iframe[src*="adframe"],
iframe[src*="mm.chitika.net"],
iframe[src*="/RealMedia/ads/"],
iframe[src*="precisionclick.com/ad/"],
iframe[src*="adsatt.espn.go.com"],
iframe[src*="rcm.amazon.com"],
iframe[src*="rcm-fr.amazon.com"],
iframe[src*="adspace"],
iframe[src*="AdPlaceholder"],
iframe[src*="ffxcam.fairfax.com.au"],
iframe[src*="contextweb.com"],
iframe[src*="ads-rm.looksmart.com"],
iframe[src*=".ru4.com"],
iframe[src*=".kanoodle.com"],
iframe[src*="adpeeps"],
iframe[src*="adbrite"],
iframe[src*="hrtrak.net"],
iframe[src*="imageshack.us/doneCode.php"],
iframe[id*="tp_frame1"],
iframe[src*="jazad.aljazeera.net"],
iframe[src*="media.nyadmcncserve"],
iframe[src*="www.google.com/pagead/ads"],
iframe[src*="reklam.gittigidiyor.com"],
iframe[src*=".specificclick.net"],
iframe[src*="optimizedby.rmxads.com"],
iframe[src*="oasis.nysun.com"],
iframe[src*=".overture.com"],
iframe[src*=".fimserve.com"],
iframe[src*="google-afc"],
iframe[src*="google-adsense"],
iframe[id*="google_ads_iframe"],
iframe[src*="werben-mit-campus-media.de/ads/"],
iframe[class*="madison_ad"],
iframe[src*=".netavenir."],
iframe[src*="bwp.versiontracker.com"],
iframe[src*="bwp.macfixit.com"],
iframe[src*="bwp.cnet.com"],
iframe[src*="bwp.news.com"],
iframe[src*=".mercuras.com"],
iframe[src*="api.gestionpub.com"],
iframe[src*=".comclick.com"],
iframe[src*="mads.com.com"],
iframe[src*="aserve.directorym.com"],
iframe[src*="connextra.com"],
iframe[src*="adaction.se"],
iframe[src*=".chitika.net"],
iframe[src*="leadmediapartners.com"],
iframe[src*="AdScroller"],
iframe[src*="pulse360.com"],
iframe[src*="/cdn/_uac/adpage.htm"],
iframe[src*="images.intellitxt.com"],
iframe[src*="box.shopoon.fr"],
iframe[class="Banner"],
iframe[id*="sponsored_links_afc"],
iframe[src*=".emediate.eu"],
iframe[src*="web.adblade.com"],
iframe[src*=".adfusion.com"],
iframe[src*=".adshuffle.com"],
iframe[name$="skybannerframe"],
iframe[src*=".quantserve.com"],
iframe[src*=".adicio.com"],
iframe[src*=".mediaforceads.com"],
iframe[src*="pub.oxado.com"],
iframe[src*="exoclick.com"],
iframe[src*="server.cpmstar.com"],
iframe[id*="cpmstar"],
iframe[class="OASAdFrame"],
iframe[src*="adreactor.com/services/advertisements/"],
iframe[src*="tracking.drive2media.com"],
iframe[id*="AdBriteIntermissionAdIframe"],

table#overtureLinksWrapper,
table.adbritetable,
table.ad_table,
table.artbannersplus,
td#tdpromo,
td[class*="blogad_"],
td[class*="adstrip"],
td[class*="AdColumn"],
td[onclick*="click.pulse360.com"],
td[class^="todayTopAd"],
td[class="midAd"],
td[id="ad_gutter"],
td[id="first_ad_unit"],
span#iTt,
span[class="advertisingctr"],
div#omniture,
div#cnnPartners,
div#gB1,
div#gB2,
div#gB3,
div.yschspns,
div#yschsec,
div#sponsorsAd,
div#leaderboardAd,
div[id="ads"],
div[id^="ads_"],
div[id*="ad440x160"],
div[id*="ad300x100"],
div[id*="ad160x600"],
div[class*="ad728x90"],
div[class*="ad300x250"],
div.werbeblock,
div.werbung,
div[id$="_ad"],
div[id$="_ads"],
div[id$="-ad"],
div.top_ad_image,
div[class*="_ad_"],
div[class^="ad_IDS"],
div[id*="swf_728x90"],
div[id*="swf_171x600"],
div[id*="swf_banner"],
div[id*="swf_400"],
div[id*="swf_160x600"],
div[id*="swf_300x250"],
div[id*="swf_171"],
div[id*="swf_adbox"],
div[id*="swf_perma_flash"],
div[id*="swf_skyscraper"],
div[id*="km_main_div"],
div#ad * object > embed[type="application/x-shockwave-flash"],
div[id*="adSpot"],
div[id*="StripAd"],
div[id^="konaLayer"][class^="konaLayer"],
div[id^="kona_"],
div[class="adPanel"],
div[id="ad-wrap"],
div#textlinks>div[class*="sponsorship"],
div#topjobs>div[class*="sponsorship"],
div[id^="ad_links"],
div[class*="adspot"],
div[id^="ad_box"],
div[class^="adbox"],
div[id="AdBlock"],
div[id^="ad_block_"],
div[id^="banner"][id$="layer"],
div[id*="OvertureSponsoredLinksBlock"],
div[id*="reklam"],
div[id*="Kanoodle"],
div[class^="adModule"],
div[id*="loweradswrapper"],
div[id="bigad"],
div[class*="context-ad"],
div[class^="adbanner"],
div[class*="_adsys_"],
div[id*="SendMessage_gads"],
div[id="box-ad"],
div[class="box-ad"],
div[id^="adxSponLink"],
div[id^="adxLeaderboard"],
div[id^="adxBigAd"],
div[id^="adxToolSponsor"],
div[class$="storyad"],
div[onclick*="click.pulse360.com"],
div[id^="adpos"],
div[id="ad"],
div[id*="adsense"],
div[class~="ad-home"],
div[id="topad_wrapper"],
div[id*="mTotem"]>iframe[id*="totem"],
div[class="advertising"],
div[class*="advertisment"],
div[id^="ad1"],
div[id^="ad2"],
div[id^="ad3"],
div[id^="ad4"],
div[id^="ad5"],
div[id^="ad6"],
div[id^="ad7"],
div[id^="ad8"],
div[id^="ad9"],
div[id="uxAd1"],
div.contextualad,
div[id^="adx_"],
div.adText,
div[class^="adsText"],
div[class*="netavenir_slideup"],
div[id="apopBg_dju"],
div[class*="apopBg"],
div[id*="advst_special"],
div[class="box"][id="dropin"],
div[class*="textAds"],
div[class~="textAd"],
div[id^="bbccom_"],
div[id^="advert_"],
div[id^="AdBroker"],
div[id^="AdSmall"],
div[id^="adSpace"],
div[id="adTicker"],
div[id="spColumnAd"],
div[class^="adBlock"],
div[class^="AdBorder"],
div[class~="adcolumn"],
div[class="adbottom"],
div[class="adwide"],
div[class="module adimg"],
div[class$="_ad"],
div[class*="publicite"],
div[id*="blogbang"],
div[id^="banner-ad"],
div[id="bannerad"],
div[class~="advertisementColumnGroup"],
div[class$="advertisement"],
div[id="rightPromoColumn"],
div[class="topAds"],
div[class="ad"],
div[class^="ad1"],
div[class="splitAd"],
div[class="wide-text-ads"],
div[class="ads"],
div[class^="adSpot"],
div[class="relatedAds"],
div[class^="adGoogle"],
div[class="adgoogle"],
div[id*="ad_google"],
div[class^="googleAdBlock"],
div[class*="google-ads"],
div[id*="google-ads-container"],
div[id="google-afc-region-one"],
div[id="google_ads_site"],
div[id*="google_ads_div"],
div[id="google_textadsBox"],
div[class*="googleads"],
div[id^="p360_ad"],
div[id^="p360-"],
div[class^="p360_"],
div[id="adjacency"],
div[id="adbar"],
div[id="dcol-sponsored"],
div[id*="sponsored_links_afc"],
div[id="sponsored-links"],
div[id^="sponLinkDiv"],
div[class="marketplace-ad"],
div[id^="ADframe"],
div[class="singleAd"],
div[id^="adsDiv"],
div[id^="AdBanner"],
div[id^="ad-front"],
div[class~="textlinkads"],
div[id*="story-sponsoredlinks"],
div[class="blk_advert"],
div[class$="bottomAd"],
div[id*="bottomad_"],
div[id="upperad"],
div[id="web_kelkoo"],
div[id^="right_ad_"],
div[class$="WerberahmenOben"],
div[class$="anzeige"],
div[class~="sb_adsW"],
div[class="sb_adsN"],
div[id="PointrollOverlay"],
div[id="blox-peel-ad"],
div[id="yahoo-contentmatch"],
div[id="rightrailads"],
div[id="right-ads"],
div[class="sponsorArea"],
div[id^="sponsorads"],
div[class="adblock"],
div[class^="yab-"][class$="-ad"],
div[class="floatedAd"],
div[id^="Liftium_"],
div[class*="SidebarAds"],
div[class*="sidebarAdblock"],
div[id*="dclkAdsDivID"],
div[class~="mod-adopenx"],
div[class="ad_header"],
div[class="ad-footer"],
div[class="ad-admob"],
div[id*="scrubbuad"],
div[class="popadtext"]

{ display: none !important; -moz-binding: none !important }

@-moz-document domain(mail.google.com) {
div#rhc>div.rh,
div.slwyWc,
div.oM { display: none !important }
}

@-moz-document domain(wunderground.com) { 
div[style*="statefarm_bg.gif"][style*="width: 300px"] { display: none !important }
}

/* turn some false positives back off */
a:link[href*="download."] img,
a:link[href*="mozilla.org"] img,
a:link[href*="netflix.com/addtoqueue"] img,
a:link[href*="click.mp3"] img,
a:link[href*="linksynergy.com"] img[src*="phobos.apple.com"],
a:link[id="AdBriteSkipThisAd"] img[src*=".adbrite.com"],
img[src*="wikipedia"],
img[src*="images.ucomics.com"],
img[src*="buttons.googlesyndication.com"],
iframe[src*="mozilla.org"],
div[id*="Player"] iframe[src*="/ads."]

{ display: inline !important }

div[class="content has_ad"] { display: block !important }

@-moz-document domain(wiki.caminobrowser.org) {
img[src*="/ad/"] { display: inline !important }
}

@-moz-document domain(apple.com) {
img[src*="/ads/"] { display: inline !important }
div[id="ad"] { display:block !important }
}

@-moz-document domain(nasdaq.com) {
div[id*="panel_content_ads"] { display: block !important }
}

@-moz-document domain(slatev.com) {
embed[type="application/x-shockwave-flash"][src*="slatev.com"] { display: inline !important }
}

@-moz-document domain(nationalgeographic.com) {
embed[type="application/x-shockwave-flash"][src*="adserver"] { display: inline !important }
}

@-moz-document domain(priceminister.com) {
a:link[href^="javascript"][href*="advert"] img { display: inline !important }
div[id="advert_list"] { display: block !important }
}

@-moz-document domain(mlb.com) {
div[id="tw_secnav_mc_ad"] { display: block !important }
}

@-moz-document domain(news.cnet.com) {
div[id="ads_backsplashSkin"] { display: block !important }
}

@-moz-document domain(lastfm.de) {
html embed[type="application/x-shockwave-flash"][width][height] { display: inline !important }
}

@-moz-document domain(macwelt.de) {
div[class*="advertisment"][id="spacer_58"] { display: block !important }
}
